
(* tsl.ML *)
(* needs updating - some bits specific to dRA *)

(* functions for producing list of rtac-type tactics from other tactics *)

(* general idea is to transform an expression like
  [ rtac th sg,
    disp_tac str sg,
    d_b_tac str tf sg,
    ... ]
  to a list of (thm, subgoal) pairs, whose meaning is the list of
  (rtac thm subgoal) tactics
  *)

(* code same as regular Isabelle version, but types less restricted *)

fun (tac1 ORELSE tac2) st =
    case Seq.pull(tac1 st) of
	None => tac2 st
      | sequencecell => Seq.make(fn()=> sequencecell);

structure no_tsl =
struct
type tactic = Tactical.tactic ;
val rtac = Tactic.rtac ;
val (op THEN) = Tactical.THEN ;
val (op THEN') = Tactical.THEN' ;
val (op THENEXP) = op THENEXP ;
val all_tac = Tactical.all_tac ;
val THENR = THENR ;
val ALLGOALS = Tactical.ALLGOALS ;
val TRY = Tactical.TRY ;
val EVERY = Tactical.EVERY ;
val disp_tac = disp_tac ;
val d_b_tac = d_b_tac ;
val d_b_s_tac = d_b_s_tac ;
val afl2tac = afl2tac ;
val glob_s_tac = glob_s_tac ;
val glob_tac = glob_tac ;
val bup_tac = bup_tac ;
val rep_glob_tac = rep_glob_tac ; 
val rep_bup_tac = rep_bup_tac ;
val tfpr = tfpr ;
val keep_tac = keep_tac ;
val weak_tac = weak_tac ;
val compSL_tac = compSL_tac ;
val compSR_tac = compSR_tac ;
end ;

structure tsl =
struct

(* these functions change the notion of a tactic;
  instead of a sequence of next states (as for an ordinary tactic),
  we use a sequence of (prepfn, next state) pairs;
  prepfn is a function which prepends (thm, subgoal) pairs to a list,
  where the pairs prepended are those which indicate the rtac steps
  which made up that use of the tactic *)
(* types of these are as for the no_tsl version, with the
  thm Seq.seq (where tactic is thm -> thm Seq.seq) replaced by
  (((thm * int) list -> (thm * int) list) * thm) Seq.seq
  where the ((thm * int) list -> (thm * int) list) is a function 
  which prepends some (thm, subgoal) pairs to a list,
  and the thm is a next state,
  ie, as well a giving a next state, each member of the sequence 
  records a list of (thm, subgoal) pairs reflecting the rtac steps
  which make up the tactic
  *)
type tactic = 
  thm -> (((thm * int) list -> (thm * int) list) * thm) Seq.seq ;
fun rtac th sg st = 
  let fun paircons nst = (Library.cons (th, sg), nst)
  in Seq.map paircons (no_tsl.rtac th sg st) end ;

fun (tac1 THEN tac2) st =
    let fun comp f1 (f2, st2) = (f1 o f2, st2) 
      fun t2 (f1, st1) = Seq.map (comp f1) (tac2 st1)
    in Seq.flat (Seq.map t2 (tac1 st)) end ;

fun (tac1 THEN' tac2) x = tac1 x THEN tac2 x;

fun all_tac state = Seq.single (I, state) ;

fun TRY tac = tac ORELSE all_tac;

fun THENR m n tf = 
  if n = m then tf m 
    else if n < m then all_tac
    else tf n THEN THENR m (n-1) tf ;

fun (tacf1 THENEXP tacf2) sg state = 
  let fun tac2 newstate = 
	(THENR sg (nprems_of newstate - nprems_of state + sg) tacf2) newstate
  in (tacf1 sg THEN tac2) state end ;

fun ALLGOALS tac st =
  let fun doall 0 = all_tac
	| doall n = tac(n) THEN doall(n-1)
  in  doall(nprems_of st)st  end;

fun EVERY tacs = BasisLibrary.List.foldr (op THEN) all_tac tacs ;

fun disp_tac str sg state =
    let val th = md2 (disp_rule str (nth (prems_of state, sg-1)))
    in rtac th sg state end ;

fun d_b_tac str tf sg state =
  let val dr = disp_rule str (nth (prems_of state, sg-1))
  in (rtac (md2 dr) THEN' tf THENEXP rtac (md1 dr)) sg state end ;

fun d_b_s_tac str plist tf sg state = 
  let val dr = disp_rule str (nth (prems_of state, sg-1))
    fun rev (1::is) sg1 = rtac (md1 dr) sg1 THEN rev is (sg1+1)
      | rev (0::is) sg1 = rev is (sg1+1)
      | rev [] _ = all_tac
  in (rtac (md2 dr) sg THEN tf sg THEN rev plist sg) state end ;

local
  fun idact sg = all_tac 
  val actcomb = op THENEXP 
  val dpact = rtac o md2 
  fun find_tm sg state = nth (prems_of state, sg-1) 
in
  val bwd_methods = {idact = idact, actcomb = actcomb,
    dpact = dpact, find_tm = find_tm} :
      (((thm * int) list -> (thm * int) list) * thm) Seq.seq meth ;
end ;

fun afl2tac afl = bs_tac bwd_methods o afl bwd_methods ;

val glob_tac = afl2tac TDaf ;
val bup_tac = afl2tac BUaf ;
fun glob_s_tac side = 
  (sidesel side (ant_tac, suc_tac) bwd_methods o TDaf bwd_methods) ;
val rep_glob_tac = 
  bs_tac bwd_methods o REPaf bwd_methods o TDaf bwd_methods ;
val rep_bup_tac = 
  bs_tac bwd_methods o REPaf bwd_methods o BUaf bwd_methods ;

fun tfpr (apr, spr) tm side = prlors rtac (sidesel side (apr, spr)) tm ;

local
fun ktes [] (Const ("DL.Sequent", _) $ L $ R) ~1 =
    rtac (exa RS exs)
  | ktes loc _ asp = d_b_tac (implode (rev loc)) (rtac (nth ([exs, exa], asp)));

fun ktaux [[]] loc tm asp = (fn i => all_tac)
  | ktaux [] loc tm asp = ktes loc tm asp
  | ktaux strs loc tm asp =
    let
      fun doch lch tm asp = 
	let val rstrs = redstrs strs lch
	in ktaux rstrs (lch::loc) tm asp end 
    in 
      case tm of
	(Const ("DL.Sequent", _) $ L $ R) =>
	  doch "|" L 0 THEN' doch "-" R 1
       | (Const ("RS.,", _) $ L $ R) => doch "l" L asp THEN' doch "r" R asp
       | (Const ("RS.;", _) $ L $ R) => doch "L" L asp THEN' doch "R" R asp
       | (Const ("RS.*", _) $ U) => doch "*" U (1-asp) 
       | (Const ("RS.@", _) $ U) => doch "@" U asp
    end ;
	
in
fun keep_tac strs sg state = ((ktaux ((map explode) strs) [] 
  (nth (prems_of state, sg-1)) ~1) THEN' glob_tac (tfpr pr_id)) sg state ;
end ;

fun weak_tac sg state = 
  case first_match (form_list (nth (prems_of state, sg-1))) of
      NONE => all_tac state
    | SOME ((_, sol1, _, _), (_, sol2, _, _)) =>
      (keep_tac [apl2str sol1, apl2str sol2] sg
	(* next step converts @($X, $Y) |- $I, $I |- *($X, $Y), etc *)
	(* was 
	THEN REPEAT (rtac sxI, Isx, bxI, Ibx] sg) *)
	THEN bup_tac (tfpr pr_dist) sg
	(* next steps converts $X, $Y |- $I or $I |- $X, $Y *)
	THEN TRY (rtac (mra RS md2 cs1) sg)
	THEN TRY (rtac (mls RS md2 ca1) sg)
	(* bup_tac needed for * @ @ * - is one enough ? *)
	THEN bup_tac (tfpr pr_dbl) sg
	THEN TRY (rtac (md1 bsA) sg)
	THEN TRY (rtac (md1 bsS) sg)
	THEN TRY (rtac (md1 blbl) sg)
	THEN TRY (rtac (md1 dcp) sg)
	THEN rtac idf sg)
      state ;

val compSL_tac = rtac (md2 ss1) THEN' rtac comps THEN' weak_tac ;
fun compSR_tac sg = rtac (md2 ss2) sg THEN rtac comps sg THEN weak_tac (sg+1) ;

end ;

(* tsltac 
  : 'a -> ('a -> (('b list -> 'b list) * 'a) Seq.seq) list -> 'b list
  where 'a is thm, 'b is thm * int
  from an initial state and a list of tsl-tactics, 
  gives a list of (thm, subgoal) pairs *)
fun tsltac st rtl = (#1 (Seq.hd (tsl.EVERY rtl st))) [] ;

(* bdev : tactic list -> unit,
  to check whether a tactic list works deterministically, 
  as required of the primitive rtac steps for
  the various functions above *)
val bdev = byev o (map DETERM) ;

val thl2rt = map (fn th => rtac th 1) ;
val tsl2rt = map (uncurry rtac) ;

(* ts2txt : thm * int -> string
  tsl2txt : (thm * int) list -> string
  produces the text of a tactic from a list of (thm, subgoal) pairs *)
fun ts2txt (th, sg) = "rtac " ^ simpnt th ^ " " ^ string_of_int sg ;
fun tsl2txt til = "byev [" ^ implode(separate",\n"(map ts2txt til)) ^ "] ;\n" ;

